// C Source File
// Created 14/04/2003; 20:59:00

#define _GENERIC_ARCHIVE
#define NO_CALC_DETECT
#include "complib_types.h"
#include "huffman.h"
#include "utils.h"
#include <tigcclib.h>         // Include All Header Files

/*===============================================================================================
== Compression KOMP
=================================================================================================
== data : adresse des donnes  compresser
== longueur : longueur des donnes  compresser
== temp : boolen faux si on va crer une variable (ajout du tag "KOMP")
                  vrai si c'est temporaire
== pb : pointeur pour renvoyer le code du probleme, si probleme il y a eu
=================================================================================================
== renvoie le handle des donnes compresses
===============================================================================================*/

HANDLE mk_komp(unsigned char *data, unsigned short longueur, unsigned char type, BOOL temp, char * pb) {
		unsigned short * stat;
		NOEUD * noeud;
		ARBRE * arbre;
		unsigned char nL;
		unsigned long l;
		unsigned char * dest;
		unsigned char * dest2;
		unsigned short long_comp;
		unsigned char decal;
		unsigned short i;
		unsigned short nbre1=0;
		unsigned short nbre2=0;
		unsigned short val;
		unsigned short mode;
		HANDLE h;
		
		*pb=COMP_NO_ERR;
		
		stat=(unsigned short *)malloc(256*sizeof(unsigned short));

		if (stat==NULL) {
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		if ((noeud=(NOEUD *)malloc(256*sizeof(NOEUD)))==NULL) {
				free(stat);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}

		nL=creation_stat(stat,data,longueur-1);

		tri_stat(stat,noeud);

		if ((arbre=(ARBRE *)malloc(nL*sizeof(ARBRE)))==NULL) {
				free(stat);
				free(noeud);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		memset(arbre,0,nL*sizeof(ARBRE));
		creation_arbre(noeud,arbre,nL);

		free(noeud);
		l=eval_long(arbre,stat);
		decal=l%8;
		
		for (i = 0 ; i < 256 ; i++) {
				val = stat[i] ;
				nbre1 += (val && val < 256) ;  //nb de (caract dont le nb est <256 et non nul)
				nbre2 += (val >= 256) ;  //nb de (caract dont le nb est >=256)
		}
		mode = ( ( nbre2 != 0) ? 2 : 1 );  //mode=2 => y'en a  + de 256   -   mode=1 => tout - de 256
		if (nbre1+2*nbre2 < 255*mode)   //est-il plus profitable de faire un mode "mix" ?
				mode = 0 ;  //oui

//  long2  => longueur des donnes compresse (en bit)
//  long2%8!=0   => un octet de plus si ca dpasse
//  1 + 255*(*stat2) + 1+255*mode+(mode==0)*(nbre1+2*nbre2+2)  : tableau de stat (longueur dpendant du mode)
		long_comp=(4+256*mode+(mode==0)*(3*nbre2+2*nbre1+2)+(l>>3)+(decal!=0)+7*(!temp));
//printf("l=%d=%d",l,longueur);
		if (long_comp>=longueur) {
				*pb=COMP_ERR_CANT_COMP;
				free(arbre);
				return H_NULL;
		}
		
		h=HeapAlloc(long_comp+2);
		if (h==H_NULL) {
				free(arbre);
				free(stat);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		
		dest=HLock(h);
		if (dest==NULL) {
				HeapFree(h);
				free(arbre);
				free(stat);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		
		if (decal==0)
				decal=8;
		*(unsigned short *)dest=long_comp;

		dest+=2;
		*(unsigned short *)dest=longueur;
		dest2=dest+long_comp-7*(!temp);

		if (ecriture_comp(data,longueur-1,arbre,dest2-1,decal-1)==0) {
				HeapFree(h);
				free(arbre);
				free(stat);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}

		dest+=2;
		*dest++ = type ;  //copie du type
		*dest++ = mode ;  //copie du mode d'arbre...
		if (mode!=0) { //mode 1 ou 2
				for (i = 0 ; i < 256 ; i++) {
						val = stat[i] ;
						*dest++ = val ;  //stockage du premier octet du nb de carct
						if (mode == 2) //si plein de nb de plus de 256
								*dest++ = val>>8 ;  //stockage du deuxime octet du nb de carct
				}
		} else {  //mode "mix"
				*dest++ = nbre2 ;  //on copie le nb de caract >256
				*dest++ = nbre1 ;  //on copie le nb de caract <256
				for (i = 0 ; i < 256 ; i++) {  //copie des grand nb
						if ((val = stat[i]) > 255) {  //c'est un gd nb
								*dest++ = i ;  //code du caract
								*dest++ = stat[i]>>8 ;  //copie de son nb
								*dest++ = val ;           //suite de la copie
						}
				}
				for (i = 0 ; i < 256 ; i++) {  //copie des peu nbreux
						val = stat[i] ;
						if (val!=0 && val < 256) {
								*dest++ = i ;  //code
								*dest++ = val ;  //copie
						}
				}
		}
		free(stat);

		if (!temp) {
				memcpy(dest2,"\0KOMP\0",6);
				*(dest2+6)=OTH_TAG;
		}
		free(arbre);

		if (HeapUnlock(h)==H_NULL) {
				HeapFree(h);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}

		return h;
}


/*===============================================================================================
== Compression komp d'une variable dans un handle temporaire
=================================================================================================
== sym_var_src : pointeur sym de la var  compresser
== pb : pointeur pour renvoyer le code du probleme, si probleme il y a eu
=================================================================================================
== renvoie le handle des donnes compresses
===============================================================================================*/

HANDLE komp_sym_temp(SYM_ENTRY * sym_var_src, char * pb) {
		unsigned char * data;
		unsigned short longueur;
		HANDLE h;
		HANDLE h_var;

		if (sym_var_src==NULL) {
				*pb=COMP_ERR_NO_SRC_VAR;
				return H_NULL;
		}
		
		h_var=sym_var_src->handle;
		data=HLock(h_var);
		if (data==NULL) {
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		longueur=*(unsigned short *)data;

		h=mk_komp(data+2,longueur,data[longueur+1],TRUE,pb);

		HeapUnlock(h_var);

		return h;
}


/*===============================================================================================
== Compression komp de donne dans une variable
=================================================================================================
== data : pointeur des donnes  dcompresser
== longueur : longueur des donnes  dcompresser
== nom_var_dest : nom de la variable dans laquelle on doit dcompresser
== str_tag : tag  ajouter  la fin
== tag_length : longueur du tag (ne se termine pas par 0! impossible d'utiliser strlen)
== pb : pointeur pour renvoyer le code du probleme, si probleme il y a eu
=================================================================================================
== renvoie le pointeur Sym de la variable cre
===============================================================================================*/

SYM_ENTRY * komp_data_var(unsigned char *data, unsigned short longueur, const char * nom_var, char * pb) {

		HANDLE h=mk_komp(data,longueur,OTH_TAG,FALSE,pb);

		if (h==H_NULL)
				return NULL;

		return creation_var(NULL,nom_var,h,pb);
}


/*===============================================================================================
== Compression komp d'une variable dans une autre variable
=================================================================================================
== sym_var_src : pointeur sym de la var  compresser
== nom_var_dest : nom de la variable dans laquelle on compresse
									si vaut NULL, compresse dans la variable de dpart
== str_tag : tag  ajouter  la fin
== tag_length : longueur du tag (ne se termine pas par 0! impossible d'utiliser strlen)
== pb : pointeur pour renvoyer le code du probleme, si probleme il y a eu
=================================================================================================
== renvoie pointeur Sym de la nouvelle variable
===============================================================================================*/

SYM_ENTRY * komp_sym_var(SYM_ENTRY * sym_var_src, const char * nom_var_dest, char * pb) {
		unsigned char * data;
		unsigned short longueur;
		HANDLE h;
		HANDLE h_var;
		
		if (sym_var_src==NULL) {
				*pb=COMP_ERR_NO_SRC_VAR;
				return H_NULL;
		}

		h_var=sym_var_src->handle;
		data=HLock(h_var);
		if (data==NULL) {
				printf("lock var fail");
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		longueur=*(unsigned short *)data;
		h=mk_komp(data+2,longueur,data[longueur+1],FALSE,pb);
		if (h==H_NULL)
				return NULL;
		HeapUnlock(h_var);
		return creation_var(sym_var_src,nom_var_dest,h,pb);

}


/*===============================================================================================
== Dcompression KOMP
=================================================================================================
== data : adresse des donnes  dcompresser
== longueur : longueur des donnes  dcompresser
== str_tag : tag  ajouter  la fin
== tag_length : longueur du tag (ne se termine pas par 0! impossible d'utiliser strlen)
== pb : pointeur pour renvoyer le code du probleme, si probleme il y a eu
=================================================================================================
== renvoie le handle des donnes dcompresses
===============================================================================================*/

HANDLE mk_dkomp(unsigned char * data, unsigned short longueur, const char * str_tag, unsigned char tag_length, char * pb) {
		ARBRE * arbre;
		unsigned short * stat;
		NOEUD * noeud;
		unsigned char nL=0;
		unsigned char * data_dest;
		unsigned char * data_comp;
		HANDLE h;
		unsigned char * n=NULL;
		unsigned char * limite;
		unsigned short i;
		unsigned short long_fichier;

		*pb=COMP_NO_ERR;

		if (strcmp(&(data[longueur-4]),"KOMP")==0 && data[3]<3) {
				*pb=COMP_ERR_NOT_A_COMP;
				return H_NULL; //ce n'est pas un komp
		}
		long_fichier=*(unsigned short *)data;
		data+=3;
		
		stat=(unsigned short *)malloc(256*sizeof(unsigned short));
		if (stat==NULL) {
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		memset(stat,0,256*sizeof(unsigned short));
		noeud=(NOEUD *)malloc(256*sizeof(NOEUD));
		if (noeud==NULL) {
				free(stat);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}

		if (*data != 0) {  //quel est le mode... 1 ou 2
				for (i = 0 ; i < 256 ; i++)
						stat[i] = data[(*data)*i+1]+(*data == 2 ? 256*data[(*data)*i+2] : 0 ) ;  //no caract=(*stat)*i   //copie du nb de caract (les deux octets !!)
		} else {  //mode "mix"
				limite = data+3+3*data[1] ;  //nombre de caract nombreux
				for (n = data+3 ; n < limite ; n += 3)  //copie des nombreux
						stat[*n] = 256*(*(n+1))+(*(n+2)) ;
				for (n = limite ; n < limite+2*data[2] ; n += 2)  //copie des peu nombreux
						stat[*n] = *(n+1) ;
		}

		for (i=0;i<256;i++)
				if (stat[i]!=0)
						nL++;  //compte des caractres diffrents

		tri_stat(stat,noeud);

		free(stat);
		
		arbre=(ARBRE *)malloc(nL*sizeof(ARBRE));
		if (arbre==NULL) {
				free(noeud);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		
		memset(arbre,0,nL*sizeof(ARBRE));
		creation_arbre(noeud,arbre,nL);
		free(noeud);

		i=long_fichier+2;
		h=HeapAlloc(i+tag_length);
		if (h==H_NULL) {
				free(arbre);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		data_dest=HLock(h);
		if (data_dest==NULL) {
				HeapFree(h);
				free(arbre);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		
		data_comp=data+1+256*(*data)+((*data)==0)*(3*data[1]+2*data[2]+2);

		n=ecriture_decomp(data_comp,long_fichier-1,data_dest+2,arbre,0,nL-2);

		*(unsigned short *)data_dest=long_fichier;
		if (tag_length==0) {
				data_dest[long_fichier+1]=*(data-1);  //copie du type
		} else {
				for (nL=0;nL<tag_length;nL++)//recopie du tag personnalis pour la dcompression data -> var
						*(n++)=*(str_tag++);
		}
		free(arbre);

		if (HeapUnlock(h)==H_NULL) {
				HeapFree(h);
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}

		return h;
}


/*===============================================================================================
== Dcompression komp de donne dans une variable
=================================================================================================
== data : pointeur des donnes  dcompresser
== longueur : longueur des donnes  dcompresser
== nom_var_dest : nom de la variable dans laquelle on doit dcompresser
== str_tag : tag  ajouter  la fin
== tag_length : longueur du tag (ne se termine pas par 0! impossible d'utiliser strlen)
== pb : pointeur pour renvoyer le code du probleme, si probleme il y a eu
=================================================================================================
== renvoie le pointeur Sym de la variable cre
===============================================================================================*/

SYM_ENTRY * dkomp_data_var(unsigned char * data, unsigned short longueur, const char * nom_var_dest,
													const char * str_tag, unsigned char tag_length, char * pb) {

		HANDLE h=mk_dkomp(data,longueur, str_tag, tag_length,pb);

		if (h==H_NULL)
				return NULL;

		return creation_var(NULL,nom_var_dest,h,pb);
}


/*===============================================================================================
== Dcompression komp d'une variable dans un handle temporaire
=================================================================================================
== sym_var_src : pointeur sym de la var  dcompresser
== pb : pointeur pour renvoyer le code du probleme, si probleme il y a eu
=================================================================================================
== renvoie le handle des donnes dcompresses
===============================================================================================*/

HANDLE dkomp_sym_temp(SYM_ENTRY * sym_var_src, char * pb) {
		unsigned char * data;
		unsigned short longueur;
		HANDLE h;
		HANDLE h_var;
		
		if (sym_var_src==NULL) {
				*pb=COMP_ERR_NO_SRC_VAR;
				return H_NULL;
		}

		h_var=sym_var_src->handle;
		data=HLock(h_var);
		if (data==NULL) {
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		longueur=*(unsigned short *)data;

		h=mk_dkomp(data+2,longueur,NULL,0,pb);

		HeapUnlock(h_var);

		return h;
}


/*===============================================================================================
== Dcompression komp d'une variable dans une variable
=================================================================================================
== sym_var_src : pointeur sym de la var  dcompresser
== nom_var_dest : nom de la variable dans laquelle on dcompresse
									si vaut NULL, dcompresse dans la variable de dpart
== pb : pointeur pour renvoyer le code du probleme, si probleme il y a eu
=================================================================================================
== renvoie le pointeur Sym de la nouvelle variable
===============================================================================================*/

SYM_ENTRY * dkomp_sym_var(SYM_ENTRY * sym_var_src, const char * nom_var_dest, char * pb) {

		unsigned char * data;
		unsigned short longueur;
		HANDLE h;
		HANDLE h_var;

		if (sym_var_src==NULL) {
				*pb=COMP_ERR_NO_SRC_VAR;
				return H_NULL;
		}

		h_var=sym_var_src->handle;
		data=HLock(h_var);
		if (data==NULL) {
				*pb=COMP_ERR_MEM;
				return H_NULL;
		}
		longueur=*(unsigned short *)data;

		h=mk_dkomp(data+2,longueur,NULL,0,pb);

		if (h==H_NULL)
				return NULL;
		HeapUnlock(h_var);
		return creation_var(sym_var_src,nom_var_dest,h,pb);

}
