// C Source File
// Created 14/04/2003; 21:06:26

#define _GENERIC_ARCHIVE
#define NO_CALC_DETECT
#include "complib_types.h"
#include <tigcclib.h>         // Include All Header Files

//const unsigned char * tab_typ[15]={"ASM","LIB","PROG","FUNC","MAT","LIST","MACR","TEXT","STR","DATA","FIG","PIC","GDB","EXPR","OTHER"};

unsigned char tab_typ2[15]={ASM_TAG,OTH_TAG,FUNC_TAG,FUNC_TAG,MATRIX_TAG,LIST_TAG,MAC_TAG,TEXT_TAG,STR_TAG,DATA_TAG,FIG_TAG,PIC_TAG,GDB_TAG,EXPR_TAG,OTH_TAG};

/*===============================================================================================
== Donne le type de la varaible cod pour filelib
=================================================================================================
== type : type selon TIOS
=================================================================================================
== renvoie le type selon filelib
===============================================================================================*/

unsigned char conv_type(unsigned char type) {
	unsigned char j=0;

	while (j<15 && type!=tab_typ2[j])
			j++;
	if (j==15)
			j--;
	return j;
}

/*===============================================================================================
== Cration d'une variable
=================================================================================================
== sym_var_src : pointeur Sym de la var de dpart
== nom_var : chaine de caractere contenant le nom de la variable
== h : handle  associer  la variable
== pb : pointeur pour renvoyer le code du probleme, si probleme il y a eu
=================================================================================================
== renvoie le pointeur Sym de la nouvelle var
===============================================================================================*/

SYM_ENTRY * creation_var(SYM_ENTRY * sym_var_src, const char * nom_var_dest, HANDLE h, char * pb) {

		BOOL arch=FALSE;
		SYM_ENTRY * SymPtr;
				
		if (nom_var_dest==NULL) {
				HeapFree(sym_var_src->handle);
				sym_var_src->handle=h;
				return sym_var_src;
		}
		
		SymPtr=DerefSym(SymFind(SYMSTR(nom_var_dest)));
		
		if (SymPtr->flags.bits.archived)//variable archive
				arch=TRUE;  //il faudra l'archiver

//on efface la var
		if (SymPtr!=NULL) {
				EM_abandon (SymPtr->handle);
				SymPtr->handle=0;
				SymDel(SYMSTR(nom_var_dest));
		}

		SymPtr=DerefSym(SymAdd(SYMSTR(nom_var_dest)));  //(re)cration de la var
		
		if (SymPtr!=NULL) {
				SymPtr->handle=h;
				
				if (arch) //la variable tait archive
						if (!EM_moveSymToExtMem(SYMSTR(nom_var_dest),HS_NULL))
								*pb=COMP_ERR_CANT_REARCH;  //can't rearchive the var
		} else {
				*pb=COMP_ERR_CANT_ADD;  //can't add the var
		}

		return SymPtr;
}

/*===============================================================================================
== Donne le type complet de la variable
=================================================================================================
== data : pointeur vers le l'adresse des donnes de la variable
=================================================================================================
== renvoie le type de la variable dans l'octet du bas
           le type de compression dans l'octet du haut
===============================================================================================*/
unsigned short get_type(unsigned char * data) {

		unsigned short longueur = *(unsigned short *)data;
		unsigned char type			= data[longueur+1];
		
		if (type==OTH_TAG) {  //peut-etre un zip
				if (strcmp(&(data[longueur-4]),"KOMP")==0)
						return COMP_TYPE_KOMP|(data[4]<<8);
				if (strcmp(&(data[longueur-3]),"ZIP")==0 && data[3]==0xFF)
						return COMP_TYPE_ZIPLIB|(tab_typ2[data[2]]<<8);
		}
		return COMP_TYPE_NORMAL|(type<<8);
}


/*===============================================================================================
== Donne le type de compression de la variable
=================================================================================================
== SymPtr : pointeur sym vers la variable
=================================================================================================
== renvoie le type de compression
===============================================================================================*/

unsigned char get_type_comp(SYM_ENTRY * SymPtr) {

		unsigned char * data		= HeapDeref(SymPtr->handle);
		unsigned short longueur = *(unsigned short *)data;
		unsigned char type			= data[longueur+1];
		
		if (type==OTH_TAG) {  //peut-etre un zip
				if (strcmp(&(data[longueur-4]),"KOMP")==0)
						return COMP_TYPE_KOMP;
				if (strcmp(&(data[longueur-3]),"ZIP")==0 && data[3]==0xFF)
						return COMP_TYPE_ZIPLIB;
		}
		return COMP_TYPE_NORMAL;
		
}


/*===============================================================================================
== Donne le type de la variable dcompresse, les cas echant le type de la variable elle-mme
=================================================================================================
== SymPtr : pointeur sym vers la variable
=================================================================================================
== renvoie le type de la variable
===============================================================================================*/

unsigned char get_type_initial_var_sym(SYM_ENTRY * SymPtr) {

		unsigned char * data		= HeapDeref(SymPtr->handle);
		unsigned short longueur = *(unsigned short *)data;
		unsigned char type			= data[longueur+1];
		
		if (type==OTH_TAG) {  //peut-etre un zip
				if (strcmp(&(data[longueur-4]),"KOMP")==0)
						return data[4];
				if (strcmp(&(data[longueur-3]),"ZIP")==0 && data[3]==0xFF)
						return tab_typ2[data[2]];
		}
		return type;
		
}


/*===============================================================================================
== Donne le type de compression de la variable
=================================================================================================
== SymPtr : pointeur sym vers la variable
=================================================================================================
== renvoie le type & ' '  si variable normale
           le type & 'K' si variable compresse avec komp
           le type & 'Z' si variable compresse avec ziplib
           types donns sous forme texte
===============================================================================================*/

unsigned short get_type_txt(SYM_ENTRY * SymPtr) {

		unsigned char * data		= HeapDeref(SymPtr->handle);
		unsigned short longueur = *(unsigned short *)data;
		unsigned char type			= data[longueur+1];
		unsigned char type_comp=0;
		unsigned char type_var=0;
		
		static unsigned char tab_typ3[]={'A','O','F','F','M','L','O','T','S','D','O','P','G','E','O'};
		
		if (type==OTH_TAG) {  //peut-etre un zip
				if (strcmp(&(data[longueur-4]),"KOMP")==0) {
						type_comp='K';
						type_var=conv_type(data[4]);
				} else if (strcmp(&(data[longueur-3]),"ZIP")==0 && data[3]==0xFF) {
						type_comp='Z';
						type_var=data[2];
				}
		} else {
				type_comp=' ';
				type_var=conv_type(type);
		}
		
		return (tab_typ3[type_var]<<8)+type_comp;
		
}